/********************************************************************
 *  Licensed Materials - Property of IBM
 *
 *  (c) Copyright IBM Corp.  2009 All Rights Reserved
 *
 *  US Government Users Restricted Rights - Use, duplication or
 *  disclosure restricted by GSA ADP Schedule Contract with
 *  IBM Corp.
 ********************************************************************/

package examples.authentication.adauthentication;

import java.io.IOException;
import java.net.Socket;
import java.util.Hashtable;

import javax.naming.Context;
import javax.naming.NamingException;
import javax.naming.directory.Attribute;
import javax.naming.directory.Attributes;
import javax.naming.directory.DirContext;
import javax.naming.directory.InitialDirContext;

import com.ibm.itim.authentication.AuthenticationFailedException;
import com.ibm.itim.authentication.ConfigurationException;
import com.ibm.itim.dataservices.model.DistinguishedName;
import com.ibm.itim.dataservices.model.ModelCommunicationException;
import com.ibm.itim.dataservices.model.ObjectNotFoundException;
import com.ibm.itim.dataservices.model.system.SystemUser;
import com.ibm.itim.dataservices.model.system.SystemUserEntity;
import com.ibm.itim.dataservices.model.system.SystemUserSearch;
import com.ibm.itim.logging.SystemLog;
import com.sun.jndi.ldap.LdapCtxFactory;

/**
 * Helper class for doing authentication
 */
public class ADAuthenticationHelper {

    private static final String DIRECTORY_UNAVAILABLE = "com.ibm.itim.authentication.DIRECTORY_UNAVAILABLE";
    private static final String ENROLE_SECURITY_CATEGORY = "examples.authentication.adauthentication";
    private static final String USER_NOT_FOUND = "com.ibm.itim.authentication.USER_NOT_FOUND";

    static final String A_TYPE = "A";
    static final String SRV_TYPE = "SRV";
    static final String CONTEXT_FACTORY = "com.sun.jndi.ldap.LdapCtxFactory";
    static final String AUTHENTICATION = "simple";
    static final String LDAP_TCP = "_ldap._tcp.";

    private DirContext ictx = null;
    public String dnsServer = null;
    private String serverHostName;
    private int serverPort;

    /** Creates new ADAuthenticationHelper */
    public ADAuthenticationHelper() {
    }

    public ADAuthenticationHelper(String dnsServer) {
        this.dnsServer = dnsServer;
    }

    /**
     * Searches for SystemUser with matching id.
     * @param uid String user id.
     * @param tenantID String giving tenantID
     * @exception AuthenticationFailedException thrown if SystemUser could not be found.
     * @exception ConfigurationException thrown if communication error detected communicating with home.
     */
    public SystemUser findUser(String uid, String tenantDN)
            throws ConfigurationException, AuthenticationFailedException {

        try {

            DistinguishedName tDN = new DistinguishedName(tenantDN);
            SystemUserSearch userSearch = new SystemUserSearch();
            SystemUserEntity user = userSearch.searchByUserID(tDN, uid);
            if (user == null) {
                SystemLog.getInstance().logWarning(
                        ENROLE_SECURITY_CATEGORY,
                        "[ADAuthenticationHelper.findUser] " + "user == null\n\t"
                                + "principal " + uid + " not found");
                throw new AuthenticationFailedException(USER_NOT_FOUND);
            }

            return (SystemUser) user.getDirectoryObject();

        } catch (ObjectNotFoundException ex) {
            SystemLog.getInstance().logError(
                    ENROLE_SECURITY_CATEGORY, ex.getMessage());
            SystemLog.getInstance().logWarning(
                    ENROLE_SECURITY_CATEGORY,
                    "[ADAuthenticationHelper.findUser] "
                            + "ObjectNotFoundException\n\t" + "principal "
                            + uid + " not found");
            throw new AuthenticationFailedException(USER_NOT_FOUND);
        } catch (ModelCommunicationException ex) {
            SystemLog.getInstance().logError(
                    ENROLE_SECURITY_CATEGORY, ex.getMessage());
            throw new ConfigurationException(DIRECTORY_UNAVAILABLE);
        }
    }

    /**
     * Check for the given domain for its existence and then check for ldap server on the given domain
     * @param domainName Name of the Domain
     * @exception ConfigurationException thrown if domain name or ldap server is failed to resolve
     */
    public void checkDomain(String domainName) throws ConfigurationException,
            AuthenticationFailedException {
        if (domainName == null || domainName.length() == 0) {
            SystemLog.getInstance().logError(
                    ENROLE_SECURITY_CATEGORY,
                    "No AD Domain name provided");
            throw new ConfigurationException("No AD Domain name provided");
        }

        // first test the sanity of the domain name itself
        try {
            SystemLog.getInstance().logInformation(
                    ENROLE_SECURITY_CATEGORY,
                    "Attempting to resolve " + domainName + " to Address record");
            ictx = createDNSLookupContext();
            Attributes attributes = ictx.getAttributes(domainName,
                    new String[] { A_TYPE });
            Attribute a = attributes.get(A_TYPE);
            if (a == null)
                throw new NamingException();
            SystemLog.getInstance().logInformation(
                    ENROLE_SECURITY_CATEGORY,
                    domainName + " resolved to " + a.get());
        } catch (NamingException e) {
            SystemLog.getInstance().logError(
                    ENROLE_SECURITY_CATEGORY,
                    "Failed to resolve " + domainName + " to Address record");
            SystemLog.getInstance().logError(
                    ENROLE_SECURITY_CATEGORY, e.getMessage());
            throw new ConfigurationException("Failed to resolve " + domainName
                    + " to A record");
        }

        // then look for the LDAP server
        String ldapServer = LDAP_TCP + domainName;
        try {
            obtainLDAPServerAndPort(ictx, domainName);
        } catch (NamingException e) {
            SystemLog.getInstance().logError(
                    ENROLE_SECURITY_CATEGORY,
                    "Failed to resolve " + ldapServer + " to SRV record");
            throw new ConfigurationException("Failed to resolve " + ldapServer
                    + " to SRV record");
        }

        // try to connect to LDAP port to make sure this machine has
        // LDAP service
        try {
            new Socket(serverHostName, serverPort).close();
        } catch (IOException e) {
            SystemLog.getInstance().logError(
                    ENROLE_SECURITY_CATEGORY,
                    "Failed to connect to LDAP: " + serverHostName + ":"
                            + serverPort);
            SystemLog.getInstance().logError(
                    ENROLE_SECURITY_CATEGORY, e.getMessage());
            throw new ConfigurationException("Failed to connect to LDAP port: "
                    + serverPort);
        }
    }

    /**
     * Provides ActiveDirectory authentication with given username,password and domain name
     * @param uid Name of the user who wants to be logged in ITIM
     * @param password Password of the user who wants to be logged in ITIM
     * @param domainName Domainname for Active Directory on which user has account
     * @exception AuthenticationFailedException thrown if user doesn.t exists or wrong password is supplied
     * @exception ConfigurationException thrown if the AuthenticationAuthority can not be constructed.
     */
    public void adAuthenticate(String uid, String password, String domainName)
            throws ConfigurationException, AuthenticationFailedException {
        if (uid == null || uid.length() == 0) {
            SystemLog.getInstance().logError(
                    ENROLE_SECURITY_CATEGORY, "No user id provided");
            throw new ConfigurationException("No user id provided");
        }

        if (uid.contains(" ")) {
            SystemLog.getInstance().logError(
                    ENROLE_SECURITY_CATEGORY,
                    "Invalid user id provided (contains spaces)");
            throw new ConfigurationException(
                    "Invalid user id provided (contains spaces)");
        }

        if (domainName == null || domainName.length() == 0) {
            SystemLog.getInstance()
                    .logError(ENROLE_SECURITY_CATEGORY,
                            "No AD domain provided");
            throw new ConfigurationException("No AD domain provided");
        }

        // bind by using the specified username/password
        Hashtable props = new Hashtable();
        props.put(Context.INITIAL_CONTEXT_FACTORY, CONTEXT_FACTORY);
        props.put(Context.SECURITY_AUTHENTICATION, AUTHENTICATION);
        props.put(Context.PROVIDER_URL,"ldap://"+serverHostName+":"+serverPort);
        props.put(Context.SECURITY_PRINCIPAL, uid);
        props.put(Context.SECURITY_CREDENTIALS, password);
        try {
            SystemLog.getInstance().logInformation(
                    ENROLE_SECURITY_CATEGORY,
                    "Ready to authenticate user " + uid);
            new InitialDirContext(props).close();
            SystemLog.getInstance().logInformation(
                    ENROLE_SECURITY_CATEGORY,
                    "Authenticated user " + uid);
        } catch (NamingException e) {
            SystemLog.getInstance().logError(
                    ENROLE_SECURITY_CATEGORY,
                    "Failed to bind to LDAP");
            SystemLog.getInstance().logError(
                    ENROLE_SECURITY_CATEGORY, e.getMessage());
            throw new AuthenticationFailedException("Either no such user '"
                    + uid + "' or incorrect password");
        }
    }

    /**
     * Create IntiaDirContext with given environment parameters
     */
    public DirContext createDNSLookupContext() throws NamingException {
        Hashtable env = new Hashtable();
        env.put(Context.INITIAL_CONTEXT_FACTORY,
                "com.sun.jndi.dns.DnsContextFactory");
        env.put(Context.PROVIDER_URL, "dns://"+dnsServer);

        return new InitialDirContext(env);
    }

    public void obtainLDAPServerAndPort(String domainName) throws NamingException {
        obtainLDAPServerAndPort(ictx, domainName);
    }

    /**
     * Use DNS and obtains the LDAP server's host name and port number.
     */
    public void obtainLDAPServerAndPort(DirContext ictx, String domainName)
            throws NamingException {
        final String ldapServer = LDAP_TCP + domainName;

        SystemLog.getInstance().logInformation(
                ENROLE_SECURITY_CATEGORY,
                "Attempting to resolve " + ldapServer + " to SRV record");
        Attributes attributes = ictx.getAttributes(ldapServer,
                new String[] { SRV_TYPE });
        Attribute a = attributes.get(SRV_TYPE);
        if (a == null) {
            throw new NamingException();
        }
        ictx.close();
        serverHostName = a.get().toString().split(" ")[3];
        serverPort = Integer.decode(a.get().toString().split(" ")[2]);
    }

    public String getDnsServer() {
        return dnsServer;
    }

    public void setDnsServer(String dnsServer) {
        this.dnsServer = dnsServer;
    }
}
